@@ -138,7 +138,11 @@ class FunctionExecutor:
             await self.destroy()
             raise
 
-    def channel(self) -> grpc.aio.Channel:
+    def channel(self) -> grpc.aio.Channel | None:
+        """Returns the gRPC channel to communicate with the Function Executor Server.
+
+        Returns None if FE is destroyed or not yet initialized.
+        """
         return self._channel
 
     def request_state_client(self) -> RequestStateClient:
@@ -148,6 +152,9 @@ class FunctionExecutor:
         return self._health_checker
 
     async def server_status(self) -> FunctionExecutorServerStatus:
+        """Returns the current status of the Function Executor Server.
+
+        Doesn't raise any exceptions."""
         return await self._server.status()
 
     async def destroy(self):
