@@ -4,7 +4,9 @@ from typing import Any
 import grpc
 
 from .client_configuration import GRPC_CHANNEL_OPTIONS
-from .function_executor_server import FunctionExecutorServer
+from .function_executor_server import (
+    FunctionExecutorServer,
+)
 
 
 class SubprocessFunctionExecutorServer(FunctionExecutorServer):
